// Licensed to the Software Freedom Conservancy (SFC) under one
// or more contributor license agreements.  See the NOTICE file
// distributed with this work for additional information
// regarding copyright ownership.  The SFC licenses this file
// to you under the Apache License, Version 2.0 (the
// "License"); you may not use this file except in compliance
// with the License.  You may obtain a copy of the License at
//
//   http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing,
// software distributed under the License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, either express or implied.  See the License for the
// specific language governing permissions and limitations
// under the License.

package org.openqa.selenium.grid.router;

import io.opentelemetry.trace.Tracer;
import org.openqa.selenium.grid.distributor.Distributor;
import org.openqa.selenium.grid.sessionmap.SessionMap;
import org.openqa.selenium.json.Json;
import org.openqa.selenium.remote.http.HttpClient;
import org.openqa.selenium.remote.http.HttpHandler;
import org.openqa.selenium.remote.http.HttpRequest;
import org.openqa.selenium.remote.http.HttpResponse;
import org.openqa.selenium.remote.http.Routable;
import org.openqa.selenium.remote.tracing.SpanDecorator;

import java.util.Objects;

import static org.openqa.selenium.remote.http.Route.combine;
import static org.openqa.selenium.remote.http.Route.get;
import static org.openqa.selenium.remote.http.Route.matching;

/**
 * A simple router that is aware of the selenium-protocol.
 */
public class Router implements Routable, HttpHandler {

  private final Routable routes;

  public Router(
    Tracer tracer,
    HttpClient.Factory clientFactory,
    SessionMap sessions,
    Distributor distributor) {
    Objects.requireNonNull(tracer, "Tracer to use must be set.");

    routes =
      combine(
        get("/status")
          .to(() -> new GridStatusHandler(new Json(), tracer, clientFactory, distributor)),
        sessions.with(new SpanDecorator(tracer, req -> "session_map")),
        distributor.with(new SpanDecorator(tracer, req -> "distributor")),
        matching(req -> req.getUri().startsWith("/session/"))
          .to(() -> new HandleSession(tracer, clientFactory, sessions)));
  }

  @Override
  public boolean matches(HttpRequest req) {
    return routes.matches(req);
  }

  @Override
  public HttpResponse execute(HttpRequest req) {
    return routes.execute(req);
  }
}
